;;; This code is made available under the Gnu Public Licence version 2
;;;
;;; Copyright John Kozak March 2005

(module timing-diagram (lib "base.ss" "jk")

  (provide parse-timing-diagram draw-timing-diagram)
  
  (require (lib "string.ss" "jk"))

  (require (lib "match.ss"))
  (require (lib "class.ss"))
  (require (lib "mred.ss" "mred"))

  (require (lib "list.ss" "srfi" "1"))

  (define *aa/symbols*
    '((#\_     . lo)
      (#\     . hi)                   ;macron; looks nicer, but not ascii
      (#\-     . hi)
      (#\\     . dn)
      (#\/     . up)
      (#\.     . xx)
      (#\Space . ??) ) )

  (define (parse-timing-diagram in)
    (let ((name     #f)
          (spec     #f)
          (tag-regs #f)
          (tags     #f)
          (diagram  #f)
          (diagrams '()) )
      (let loop ()
        (let ((line (read-line in)))
          (regex-case line
            (eof? (reverse diagrams))
            ("name: ([A-Za-z0-9-]+)" =>
             (lambda (n)
               (set! name n)
               (loop) ) )
            ("spec: ([A-Za-z0-9-]+)" =>
             (lambda (s)
               (set! spec s)
               (loop) ) )
            ("timing: +([0-9]+)->([0-9]+) +([0-9]+) +([0-9]+)" =>
             (lambda (t1 t2 t-min t-max)
               (loop) ) )
            ("[ ]+\\|([ 0-9]+)\\| *" =>
             (lambda (d)
               (when (not tag-regs)
                 (set! tag-regs (make-vector (string-length d) 0)) )
               (for-each (lambda (i)
                           (let ((c (string-ref d i)))
                             (when (not (equal? c #\Space))
                               (update! (vector-ref tag-regs i) (&_ * 10))
                               (update! (vector-ref tag-regs i) (&_ + (- (char->integer c) (char->integer #\0)))) ) ) )
                         (iota (string-length d)) ) 
               (loop) ) )
            ("([<>]) *([a-z0-9*-]+)[ ]*\\|(.+)\\|" =>
             (lambda (w n d)
               (when (not diagram)
                 (set! diagram '())
                 (set! tags (list->vector (let loop ((i 0) (ans '()))
                                            (if (= i (vector-length tag-regs))
                                                ans
                                                (loop (1+ i) (if (zero? (vector-ref tag-regs i))
                                                                 ans
                                                                 (append ans (list i)) ))  ) )))
                 (set! tag-regs #f) )
               (update! diagram (_& cons `(,(string->symbol n)
                                           ,(string->symbol w)
                                           ,(list->vector (map (lambda (x)
                                                                 (cond ((assoc x *aa/symbols*) => cdr)
                                                                       (else (error "??: ~S" x)) ) )
                                                               (string->list d) ))) ))
               (loop) ) )
            (""
             (when (and name diagram)
               (update! diagrams (_& cons `(,name ,spec ,tags ,(reverse diagram))))
               (set! name    #f)
               (set! spec    #f)
               (set! tags    #f)
               (set! diagram #f) )
             (loop) )
            ("(.*)" =>
             (lambda (x)
               (error "can't match: ~S" x) ) ) ) ) ) ) )

  (define *black*      (make-object color%   0   0   0))
  (define *dark-gray*  (make-object color% 128 128 128))
  (define *light-gray* (make-object color% 192 192 192))
  (define *yellow*     (make-object color% 255 215   0))

  (define (draw-timing-diagram td dc . rest)
    ;; the various magic nimbers below should be derived from font metrics
    (let* ((x-start           (if (null? rest) 0 (car rest)))
           (y-start           (if (null? rest) 0 (cadr rest)))
           (font              (make-object font% 20 'default 'normal 'normal #f 'default))
           (extra-line-space  10)
           (transition-width  20)
           (signal-height     #f)
           (signal-name-width 0)
           (y-signal          (+ 20 y-start))
           (choose-pen        (lambda (colour width style)
                                (send dc set-pen (send the-pen-list find-or-create-pen colour width style)) ))
           (choose-brush      (lambda (colour style)
                                (send dc set-brush (send the-brush-list find-or-create-brush colour style)) )) )
      (send dc set-font font)
      (match td
        (`(,name ,spec ,tracez ,wires)
         
         ;; draw signal labels (and determine width of that box)
         (for-each (lambda (wire i)
                     (match wire
                       (`(,signal ,direction ,transitionz)
                        (let ((name       (symbol->string signal))
                              (colour     (make-object color% 0 0 0))
                              (active-low #f) )
                          (when (string-starts-with? name "-")
                            (set!    active-low #t)
                            (update! name       (&_ substring 1)) )
                          (when (string-ends-with? name "*")
                            (set!    colour (make-object color% 0 255 0))
                            (update! name   (&__ substring 0 (1- (string-length name)))) )
                          (update! name string-upcase)
                          (send dc set-text-foreground colour)
                          (choose-pen colour 3 'solid)
                          (let-values (([width height descent space] (send dc get-text-extent name)))
                            (set! signal-height height) ;N.B. wasteful
                            (let ((y (+ (* i (+ height extra-line-space)) extra-line-space y-signal)))
                              (when (> width signal-name-width)
                                (set! signal-name-width width) )
                              (when active-low
                                (send dc draw-line x-start y (+ x-start width) y) )
                              (send dc draw-text name x-start y #f 0 0) ) ) ) ) ) )
                   wires (iota (length wires)) )

         ;; draw the signals
         (for-each (lambda (wire i)
                     (match wire
                       (`(,signal ,direction ,transitionz)
                        (let ((y (+ (* i (+ signal-height extra-line-space)) extra-line-space y-signal))
                              (x (+ x-start signal-name-width transition-width)) )
                          (for-each (lambda (t j)
                                      (let ((x1 (+ x (* transition-width j))))
                                        (choose-pen *black* 1 'solid)
                                        (case t
                                          ((hi)
                                           (send dc draw-line x1 y (+ x1 transition-width) y) )
                                          ((lo)
                                           (send dc draw-line x1 (+ y signal-height)
                                                 (+ x1 transition-width) (+ y signal-height)) )
                                          ((up)
                                           (send dc draw-line x1 (+ y signal-height)
                                                 (+ x1 transition-width) y) )
                                          ((dn)
                                           (send dc draw-line x1 y 
                                                 (+ x1 transition-width) (+ y signal-height) ) )
                                          ((??)
                                           (choose-pen *black* 0 'transparent)
                                           (cond
                                            ((and (< 0 j) (eq? (vector-ref transitionz (1- j)) 'xx))
                                             (choose-brush *dark-gray* 'solid)
                                             (send dc draw-polygon
                                                   `(,(make-object point% x1 y)
                                                     ,(make-object point% x1 (+ y signal-height))
                                                     ,(make-object point% (+ x1 (/ transition-width 2)) (+ y (/ signal-height 2))) ) )
                                             (choose-brush *light-gray* 'solid)
                                             (send dc draw-polygon
                                                   `(,(make-object point% (+ x1 (/ transition-width 2)) (+ y (/ signal-height 2)))
                                                     ,(make-object point% (+ x1 transition-width) y)
                                                     ,(make-object point% (+ x1 transition-width) (+ y signal-height)) ) ) )
                                            (else
                                             (choose-brush *light-gray* 'solid)
                                             (send dc draw-rectangle x1 y transition-width signal-height) ) ) )
                                          ((xx)
                                           (choose-pen *black* 0 'transparent)
                                           (cond
                                            ((and (< 0 j) (eq? (vector-ref transitionz (1- j)) '??))
                                             (choose-brush *dark-gray* 'solid)
                                             (choose-brush *light-gray* 'solid)
                                             (send dc draw-polygon
                                                   `(,(make-object point% x1 y)
                                                     ,(make-object point% x1 (+ y signal-height))
                                                     ,(make-object point% (+ x1 (/ transition-width 2)) (+ y (/ signal-height 2))) ) )
                                             (choose-brush *dark-gray* 'solid)
                                             (send dc draw-polygon
                                                   `(,(make-object point% (+ x1 (/ transition-width 2)) (+ y (/ signal-height 2)))
                                                     ,(make-object point% (+ x1 transition-width) y)
                                                     ,(make-object point% (+ x1 transition-width) (+ y signal-height)) ) ) )
                                            (else
                                             (choose-brush *dark-gray* 'solid)
                                             (send dc draw-rectangle x1 y transition-width signal-height) ) ) ) ) ) )
                                    (vector->list transitionz)
                                    (iota (vector-length transitionz)) ) ) ) ) ) 
                   wires (iota (length wires)) )

         ;; draw the vertical transition traces
         (let ((y0 y-signal)
               (y1 (+ y-signal (* (+ signal-height extra-line-space) (length wires)))) )
           (choose-pen *yellow* 1 'short-dash)
           (send dc set-text-foreground *yellow*)
           (for-each (lambda (trace i)
                       (let ((x    (+ x-start signal-name-width transition-width (* transition-width (+ trace 1/2))))
                             (name (format "~a" i)) )
                         (let-values (([width height descent space] (send dc get-text-extent name)))
                           (send dc draw-text name (- x (/ width 2)) (- y0 signal-height)) )
                         (send dc draw-line x y0 x y1) ) )
                     (vector->list tracez)
                     (iota (vector-length tracez)) ) )
         
         (send dc set-text-foreground *black*)
         (choose-pen *black* 1 'solid)
         
         (void) ) ) ) )
  
  (define (make-timing-diagram-in-frame td)
    (let* ((frame  (instantiate frame% ((format "~a: ~a" (cadr td) (car td)))))
           (canvas (instantiate canvas% (frame)
                     (paint-callback (lambda (canvas dc)
                                       (draw-timing-diagram td dc 100 100) )) )) )
      (send frame show #t)
      (yield 'wait) ) )

)